 /**************************************************************************
  *                                                                        *
  *               Copyright (C) 1996, Silicon Graphics, Inc.               *
  *                          All Rights Reserved.                          *
  *                                                                        *
  *  The files in this subtree  contain  UNPUBLISHED  PROPRIETARY  SOURCE  *
  *  CODE of Silicon Graphics, Inc.;  the  contents  of  these  files may  *
  *  not be disclosed to third  parties,  copied  or  duplicated  in  any  *
  *  form, in whole or in part,  without  the  prior  written  permission  * 
  *  of  Silicon Graphics, Inc.                                            *
  *                                                                        *
  *  RESTRICTED RIGHTS LEGEND:                                             *
  *  Use,  duplication  or  disclosure  by  the  Government is subject to  *
  *  restrictions as set forth in subdivision (c)(1)(ii) of the Rights in  *
  *  Technical Data and Computer Software clause at  DFARS  252.227-7013,  *
  *  and/or in similar or successor  clauses in the FAR,  DOD or NASA FAR  *
  *  Supplement.  Unpublished - rights reserved  under the Copyright Laws  *
  *  of the United States.                                                 *
  *                                                                        *
  *  THIS SOFTWARE IS  PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,  *
  *  EXPRESS,  IMPLIED OR  OTHERWISE,  INCLUDING  WITHOUT LIMITATION, ANY  *
  *  WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.      *
  *                                                                        *
  *  IN  NO  EVENT  SHALL  SILICON  GRAPHICS  BE  LIABLE FOR ANY SPECIAL,  *
  *  INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF  ANY  KIND,  OR ANY  *
  *  DAMAGES WHATSOEVER RESULTING FROM LOSS  OF  USE,  DATA  OR  PROFITS,  *
  *  WHETHER OR NOT ADVISED OF  THE  POSSIBILITY  OF  DAMAGE,  AND ON ANY  *
  *  THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR  *
  *  PERFORMANCE OF THIS SOFTWARE.                                         *
  **************************************************************************/

/*
    Author : Patrick Bouchaud
             galaad@neu.sgi.com
*/

#define	MAIN_PROGRAM

#include <sys/types.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <values.h>
#include <sys/mman.h>

#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>

#include "extractor.h"
#include "manipulator.h"
#include "rgba.h"
#include "mallocbis.h"

#define	INF(a,b) ((a)<(b) ? (a) : (b))
#define	SUP(a,b) ((a)>(b) ? (a) : (b))

#define	OUTLINE_2D_TEXTURE	1
#define	OUTLINE_3D_TEXTURE	2
#define	GLOBAL_TEXTURE		3

static int
	textureMode		= GL_NEAREST,
	minfilter		= GL_LINEAR_MIPMAP_LINEAR,
	polygonmode		= GL_FALSE,
	capping			= GL_TRUE,
	outtextured		= GL_FALSE,
	outlined		= GL_TRUE,
	antialiased		= GL_TRUE,
	hiddenlines		= GL_FALSE,
	polygonoffset	= GL_TRUE;

static	float *RGBA = NULL;
static float rotationMatrix[16];
static PlaneStructure clipplane;

static int minval=255, maxval=0, GISwin, redraw=1;
static int update, updateDList, updateTLUT;
static int W, H, spinning=0, dimx, dimy, dimz;
static float maxdim, xstep, ystep, zstep;
static float nearclip, farclip;
static struct { float x,y,z; } obs, centre;
static struct {
	int		lastx, lasty,
			middledown, rightdown, leftdown;
	float	dx, dy;
} mouse;

static Extractor extractor;
unsigned char *scalar, *voxel;

typedef struct _segment {
    short   i, j, k, which, last;
    struct  _segment *next;
} Segment;

typedef struct _contour {
    Segment *firstSegment;
    struct _contour *next;
} Contour;

Contour *contourList = NULL;
Segment *segmentList = NULL;
#define newContour( contour ) newItem( contour, contourList, Contour )
#define newSegment( segment ) newItem( segment, segmentList, Segment )
#define freeContour( contour ) freeItem( contour, contourList )
#define freeSegment( segment ) freeItem( segment, segmentList )

static Contour *firstContour = NULL, *contour;
static Segment *firstSegment = NULL, *segment;

static void freeContours( void );
static void drawContours();
static void drawTextured( int );
static void setCappedPlane( void );
static void drawCappedPlane( void );
static void drawOutTextured( void );
static void drawOutlines( void );

static void Redraw( void );
static void redisplay( void );
static void Mouse( int, int, int, int );
static void Motion( int, int );
static void Reshape( int, int );
static void Key( unsigned char, int, int );
static int ReadData( char * );
static void setSceneBBox( float, float, float, float, float, float );

int isValid( int, int, int );	/* Extractor's Callbacks */
void bgnContour( void );
void endContour( void );
void setContour( int, int, int, int, int );
void newLUTCallBack( int * );
void newPlaneCallBack( void );
void menuCallBack( int );

static void bindTexture( int );
static void loadTLUT( void );

/*---------------------------------------------------------------------------*/
/*				MAIN
/*---------------------------------------------------------------------------*/
void
main( int argc, char *argv[] )
{
	glutInit( &argc, argv );

/*
	Local Initializations
	---------------------
*/
	if ( argc < 5 )
	{
		fprintf(stderr, "\nUSAGE: Extractor filename dimx dimy dimz [D]\n");
		exit(-1);
	}
	dimx = atoi( argv[2] );
	dimy = atoi( argv[3] );
	dimz = atoi( argv[4] );
	ReadData( argv[1] );
	setSceneBBox( 0., (float)dimx, 0., (float)dimy, 0., (float)dimz );
	
/*
	Graphics Initializations
	------------------------
*/
	glutInitDisplayMode( GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH | GLUT_STENCIL );
	glutInitWindowPosition( 0, 0 );
	glutInitWindowSize( 850, 1000 );
	GISwin = glutCreateWindow( "G.I.S. Extraction" );

	glutDisplayFunc( Redraw );
	glutReshapeFunc( Reshape );
	glutMouseFunc( Mouse );
	glutIdleFunc(redisplay);
	glutMotionFunc( Motion );
	glutKeyboardFunc(Key);

	extractor.ifirst = 0; extractor.inum = dimx;
	extractor.jfirst = 0; extractor.jnum = dimy;
	extractor.kfirst = 0; extractor.knum = dimz;

	extractor.testFunc = isValid;
	extractor.bgnFunc = bgnContour;
	extractor.endFunc = endContour;
	extractor.setFunc = setContour;

	newLUTCallBack( NULL );

	if (argc > 5)
	{
		RGBA = (float *)Default_RGBA;
		YACME_init(
			900, 471, 400, 545,
			14, &RGBA,
			newLUTCallBack
		);
	}
	else
	{
		outlined = GL_FALSE;
		YACME_init(
			900, 471, 400, 545,
			256, &RGBA,
			newLUTCallBack
		);
	}
	InitManipDisplay( 900, 0, 400, 400, newPlaneCallBack );
	glutMainLoop();
}

/*---------------------------------------------------------------------------*/
/*	InitGraphique
/*---------------------------------------------------------------------------*/
static void
InitGraphique( void )
{
	makePopupMenu();

	glEnable( GL_DEPTH_TEST );
	glDepthFunc( GL_LEQUAL );
	
	glShadeModel( GL_FLAT );

	glClearColor( 0., 0., 0., 1. );
	glClearStencil( 0 );
	glClearDepth( 1. );
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
	glutSwapBuffers();

	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();
	glRotatef( 180., 1., 0., 0. );
	glGetFloatv( GL_MODELVIEW_MATRIX, rotationMatrix );
}

/*---------------------------------------------------------------------------*/
/*	newPlaneCallBack
/*---------------------------------------------------------------------------*/
void
newPlaneCallBack( void )
{
	if (capping)
	{
		redraw = 1;
		glutIdleFunc( redisplay );
	}
}

/*---------------------------------------------------------------------------*/
/*	newLUTCallBack
/*---------------------------------------------------------------------------*/
typedef void (*voidFuncPtr)(void);

void
newLUTCallBack( int *hasChanged )
{
	if (hasChanged != NULL)
	{
		/* update only for ALPHA changed */
		update = hasChanged[3];
	}
	else update = 1;
	if (update)
	{
		glutIdleFunc( redisplay );
	}
	updateTLUT = GL_TRUE;
	redraw = 1;
}

/*---------------------------------------------------------------------------*/
/*	setSceneBBox
/*---------------------------------------------------------------------------*/
static void
setSceneBBox(
    float xmin, float xmax,
    float ymin, float ymax,
    float zmin, float zmax
)
{
	maxdim = xmax - xmin;
	maxdim = SUP( maxdim, (ymax-ymin) );
	maxdim = SUP( maxdim, (zmax-zmin) );

	centre.x = (xmin+xmax)/2.;
	centre.y = (ymin+ymax)/2.;
	centre.z = (zmin+zmax)/2.;

	obs.x = 0.;
	obs.y = 0.;
	obs.z = 1.5*maxdim;

	xstep = .01 * (xmax - xmin);
	ystep = .01 * (ymax - ymin);
	zstep = .05 * (zmax - zmin);
}

/*---------------------------------------------------------------------------*/
/*	Redraw
/*---------------------------------------------------------------------------*/
static void
Redraw( void )
{
	static float Rx, Ry;

	if (mouse.leftdown)
	{
		if (mouse.middledown)
			obs.z -= zstep*((float) mouse.dx);
		else
		{
			obs.x -= xstep*((float) mouse.dx );
			obs.y += ystep*((float) mouse.dy );
		}
	}
	else if (mouse.middledown)
	{
		Ry = mouse.dx;
		Rx = mouse.dy;
	}

	if (update)
	{
		update = 0;
		freeContours();
		ExtractContour( &extractor );
		updateDList = 1;
	}
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

	nearclip = SUP( obs.z-maxdim, 1./maxdim );
	farclip = obs.z + maxdim;
	glMatrixMode( GL_PROJECTION);
	glLoadIdentity();
#define	DIM nearclip*1.732
	glOrtho( -DIM, DIM, -DIM, DIM, nearclip, farclip );
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glTranslatef( -obs.x, -obs.y, -obs.z );
	if (spinning && ((Rx!=0.) || (Ry!=0.)))
	{
		glPushMatrix();
		glLoadIdentity();
		glRotatef( Ry, 0., 1., 0. );
		glRotatef( Rx, 1., 0., 0. );
		glMultMatrixf( rotationMatrix );
		glGetFloatv( GL_MODELVIEW_MATRIX, rotationMatrix );
		glPopMatrix();
	}
	glMultMatrixf( rotationMatrix );
	glTranslatef( -centre.x, -centre.y, -centre.z );

	if (capping)
	{
		glDisable( GL_CULL_FACE );
		setCappedPlane();
	}
	else
	{
		glEnable( GL_CULL_FACE );
		glCullFace( GL_BACK );
	}

	if (outtextured)
		drawOutTextured();
	else if (hiddenlines)
	{
		glColorMask( GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE );
		drawTextured( GL_FALSE );
		glColorMask( GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE );
	}
	else
	{
#if 1
		drawContours();	/* with colors */
#else
		drawTextured( GL_TRUE );
		glDisable( GL_TEXTURE_3D_EXT );
		glDisable( GL_TEXTURE_COLOR_TABLE_SGI );
#endif
	}

	if (outlined && !outtextured)
		drawOutlines();

	if (capping )
		drawCappedPlane();

/*
	undo translation only (rotation is incremental)
*/
	glTranslatef( obs.x, obs.y, obs.z );
	glutSwapBuffers();
	redraw = 0;
}

/*---------------------------------------------------------------------------*/
/*	redisplay
/*---------------------------------------------------------------------------*/
static void
redisplay( void )
{
	int curwin = glutGetWindow();
	if ((!curwin) || (curwin == GISwin) || redraw)
	{
		glutSetWindow( GISwin );
		glutPostRedisplay();
	}
}

/*---------------------------------------------------------------------------*/
/*	Mouse
/*---------------------------------------------------------------------------*/
static void
Mouse( int button, int state, int x, int y )
{
	mouse.lastx = x;
	mouse.lasty = y;
	mouse.dx = 0.;
	mouse.dy = 0.;
	switch (button)
	{
		case GLUT_LEFT_BUTTON:
			mouse.leftdown = (state==GLUT_DOWN);
			spinning = 0;
			break;

		case GLUT_RIGHT_BUTTON:
			mouse.rightdown = (state==GLUT_DOWN);
			spinning = 0;
			break;

		case GLUT_MIDDLE_BUTTON:
			mouse.middledown = (state==GLUT_DOWN);
			if (mouse.middledown) spinning = 1;
			break;
	}
}

/*---------------------------------------------------------------------------*/
/*	Motion
/*---------------------------------------------------------------------------*/
static void
Motion( int x, int y )
{
	mouse.dx = x-mouse.lastx,
	mouse.dy = y-mouse.lasty;
	mouse.lastx = x;
	mouse.lasty = y;
	redraw = mouse.middledown || mouse.leftdown;
}

/*---------------------------------------------------------------------------*/
/*		Reshape
/*---------------------------------------------------------------------------*/
static void
Reshape( int w, int h )
{
	static int first = 1;
	glViewport( 0, 0, w, h );
	W = w;
	H = h;
	mouse.leftdown = 0;
	mouse.middledown = 0;
	mouse.rightdown = 0;
	redraw = 1;

	if (first)
	{
		first = GL_FALSE;
		InitGraphique();
	}
}

/*---------------------------------------------------------------------------*/
/*		Key
/*---------------------------------------------------------------------------*/
static void
Key(unsigned char key, int x, int y)
{
	switch (key)
	{
		case 'A':
		case 'a': antialiased = !antialiased; break;

		case 'b':
		case 'B': polygonmode = !polygonmode; break;

		case 'C':
		case 'c': capping = !capping; break;

		case 'H':
		case 'h':
			hiddenlines = !hiddenlines;
			if (hiddenlines) outlined = 1;
			break;

		case 'O':
		case 'o':
			if (!hiddenlines) outlined = !outlined;
			if (!outlined) outtextured = 0;
			break;

		case 'P':
		case 'p': polygonoffset = !polygonoffset; break;

		case 'T':
		case 't': if (outlined) outtextured = !outtextured; break;


		case 27: /* Escape */
			exit(0);
	}
	redraw = 1;
}

/*---------------------------------------------------------------------------*/
/*			ReadData
/*---------------------------------------------------------------------------*/
static int
ReadData( char *filename )
{
	int		fd;

	if ((fd=open( filename, O_RDONLY )) < 0)
	{
		fprintf( stderr, "ERROR::ReadData:: unable to open file\n");
		exit(-1);
	}
	scalar = (char *) malloc( dimx*dimy*dimz );
	read( fd, scalar, dimx*dimy*dimz );
	close( fd );
}

/*===========================================================================*/
/*	Extractor's Callbacks
/*	used to generate a linked list of "Contours",
/*	which are drawn each frame by the drawContours() routine
/*===========================================================================*/
/*
	isValid
	-------
	This function is called by the extractor to determine whether the
	cell located at (i,j,k) is valid or not (for instance it's part of
	a well, or has been invalidated by the user)
	in any case, the associated alpha value has been set to 0
*/
int
isValid( int i, int j, int k )
{
	int ndx;

	if ((i<extractor.ifirst || i>=extractor.inum) ||
		(j<extractor.jfirst || j>=extractor.jnum) ||
		(k<extractor.kfirst || k>=extractor.knum))
		return 0;

	ndx = scalar[i + dimx*( j + dimy*k )];
	return (RGBA[4*ndx+3] != 0);
}

/*
	bgnContour, endContour
	----------------------
*/
void
bgnContour( void )
{
	if (firstContour == NULL)
	{
		newContour( firstContour );
		contour = firstContour;
	}
	else
	{
		newContour( contour->next );
		contour = contour->next;
	}
	contour->next = NULL;

	contour->firstSegment = NULL;
}
void endContour( void ) {;}

/*
	setContour
	----------
*/
void
setContour( int which, int i, int j, int k, int last )
{
	int count;
	if (contour->firstSegment == NULL)
	{
		newSegment( contour->firstSegment );
		segment = contour->firstSegment;
	}
	else
	{
		newSegment( segment->next );
		segment = segment->next;
	}
	segment->next = NULL;
	segment->which = which;
	segment->i = i;
	segment->j = j;
	segment->k = k;
	segment->last = last;
}

/*---------------------------------------------------------------------------*/
/*	freeContours
/*---------------------------------------------------------------------------*/
void
freeContours( void )
{
	Contour *c, *nextContour;
	Segment *s, *nextSegment;
	for (c=firstContour; c!=NULL; c=nextContour)
	{
		nextContour = c->next;
		for (s=c->firstSegment; s!=NULL; s=nextSegment)
		{
			nextSegment = s->next;
			freeSegment( s );
		}
		freeContour( c );
	}
	firstSegment = NULL;
	firstContour = NULL;
}

/*---------------------------------------------------------------------------*/
/*	drawContours
/*---------------------------------------------------------------------------*/
void
drawContours()
{
	int count, ndx;
	Segment *s;
	Contour *c;
	static int dlist = 0;

#ifdef DLIST
	if (!dlist)
		dlist = glGenLists( 1 );
	if ( updateDList )
		glNewList( dlist, GL_COMPILE_AND_EXECUTE );
	else
	{
		glCallList( dlist );
		return;
	}
#endif

#undef DRAWIT
#define DRAWIT( i0, j0, k0, i1, j1, k1, num, change )\
    count = num;\
    while (count--)\
    {\
        glVertex3i( i0, j0, k0 );\
        glVertex3i( i1, j1, k1 );\
		if (!hiddenlines)\
		{\
			ndx = scalar[i+dimx*(j+dimy*k)];\
        	glColor4fv( &RGBA[4*ndx] );\
		}\
        change;\
    }

#undef DRAWIT_CLOSED
#define DRAWIT_CLOSED( i0, j0, k0, i1, j1, k1, num, change )\
	DRAWIT( i0, j0, k0, i1, j1, k1, num, change )\
	glVertex3i( i0, j0, k );\
	glVertex3i( i1, j1, k );

	for (c=firstContour; c!=NULL; c=c->next)
	{
		int i, j, k, last;
		glBegin( GL_QUAD_STRIP );
		for (s=c->firstSegment; s!=NULL; s=s->next)
		{
			i=s->i, j=s->j, k=s->k, last=s->last;
			switch (s->which)
			{
				case LEFT:
					DRAWIT( i, j, k, i+1, j, k, last-k+1, k++ );
					break;
		
				case RIGHT:
					DRAWIT( i, j+1, k+1, i+1, j+1, k+1, k-last+1, k-- );
					break;
		
				case TOP:
					DRAWIT( i, j, k+1, i+1, j, k+1, last-j+1, j++ );
					break;
		
				case BOTTOM:
					DRAWIT( i, j+1, k, i+1, j+1, k, j-last+1, j-- );
					break;
		
				case FRONT:
					DRAWIT_CLOSED( i+1, j, k, i+1, j+1, k, last-k+1, k++ );
					break;
		
				case BACK:
					DRAWIT_CLOSED( i, j+1, k, i, j, k, last-k+1, k++ );
					break;
			}
		}
/*
		close the contour
		-----------------
*/
		s = c->firstSegment;
		if (s->which == LEFT)
		{
			DRAWIT( s->i, s->j, s->k, s->i+1, s->j, s->k, 1, k++ );
		}
		glEnd();
	}

#ifdef DLIST
	if ( updateDList )
		glEndList();
	updateDList = 0;
#endif
}

/*---------------------------------------------------------------------------*/
/*	drawTextured
/*---------------------------------------------------------------------------*/
void
drawTextured( int textured )
{
	int count, ndx;
	Segment *s;
	Contour *c;
	static int dlist = 0;

	if (textured)
		bindTexture( GLOBAL_TEXTURE );

	glColor3f( 1., 1., 1. );

#ifdef DLIST
	if (!dlist)
		dlist = glGenLists( 1 );
	if ( updateDList )
		glNewList( dlist, GL_COMPILE_AND_EXECUTE );
	else
	{
		glCallList( dlist );
		return;
	}
#endif

	for (c=firstContour; c!=NULL; c=c->next)
	{
		glBegin( GL_QUADS );
		for (s=c->firstSegment; s!=NULL; s=s->next)
		{
			int i=s->i, j=s->j, k=s->k, last=s->last;
			switch (s->which)
			{
				case LEFT:
					if (textured) glTexCoord3i( i, j, k );
					glVertex3i( i, j, k );
					glVertex3i( i+1, j, k );

					if (textured) glTexCoord3i( i, j, last+1 );
					glVertex3i( i+1, j, last+1 );
					glVertex3i( i, j, last+1 );
					break;

				case RIGHT:
					if (textured) glTexCoord3i( i, j, k+1 );
					glVertex3i( i, j+1, k+1 );
					glVertex3i( i+1, j+1, k+1 );

    				if (textured) glTexCoord3i( i, j, last );
					glVertex3i( i+1, j+1, last );
					glVertex3i( i, j+1, last );
					break;

				case TOP:
					if (textured) glTexCoord3i( i, j, k );
					glVertex3i( i, j, k+1 );
					glVertex3i( i+1, j, k+1 );

	    			if (textured) glTexCoord3i( i, last+1, k );
   		 			glVertex3i( i+1, last+1, k+1 );
   		 			glVertex3i( i, last+1, k+1 );
					break;

				case BOTTOM:
					if (textured) glTexCoord3i( i, j+1, k );
					glVertex3i( i, j+1, k );
					glVertex3i( i+1, j+1, k );

	    			if (textured) glTexCoord3i( i, last, k );
   		 			glVertex3i( i+1, last, k );
   		 			glVertex3i( i, last, k );
					break;

				case FRONT:
					if (textured) glTexCoord3i( i, j, k );
					glVertex3i( i+1, j, k );
					glVertex3i( i+1, j+1, k );

					if (textured) glTexCoord3i( i, j, last+1 );
					glVertex3i( i+1, j+1, last+1 );
					glVertex3i( i+1, j, last+1 );
					break;
		
				case BACK:
					if (textured) glTexCoord3i( i, j, k );
					glVertex3i( i, j+1, k );
					glVertex3i( i, j, k );

					if (textured) glTexCoord3i( i, j, last+1 );
					glVertex3i( i, j, last+1 );
					glVertex3i( i, j+1, last+1 );
					break;
			}
		}
		glEnd();
	}

#ifdef DLIST
	if ( updateDList )
		glEndList();
	updateDList = 0;
#endif
}

/*---------------------------------------------------------------------------*/
/*	drawOutTextured
/*---------------------------------------------------------------------------*/
#define TEXX	32
#define TEXY	32
#define TEXZ	32
void
drawOutTextured( void )
{
	int count, ndx, texcount;
	Segment *s;
	Contour *c;
	static GLuint texobj;

	static int first = GL_TRUE;

	bindTexture( OUTLINE_2D_TEXTURE );
		
#undef DRAWIT
#define DRAWIT( i0, j0, k0, i1, j1, k1, num, change )\
    count = num;\
    while (count--)\
    {\
		glTexCoord2i(0,texcount); glVertex3i( i0, j0, k0 );\
        glTexCoord2i(1,texcount); glVertex3i( i1, j1, k1 );\
		texcount++;\
		if (!hiddenlines)\
		{\
			ndx = scalar[i+dimx*(j+dimy*k)];\
        	glColor4fv( &RGBA[4*ndx] );\
		}\
        change;\
    }

#undef DRAWIT_CLOSED
#define DRAWIT_CLOSED( i0, j0, k0, i1, j1, k1, num, change )\
	DRAWIT( i0, j0, k0, i1, j1, k1, num, change )\
	glTexCoord2i(0,texcount); glVertex3i( i0, j0, k );\
	glTexCoord2i(0,texcount); glVertex3i( i1, j1, k );\
	texcount++;

	for (c=firstContour; c!=NULL; c=c->next)
	{
		int i, j, k, last;
		texcount = 0;
		if (hiddenlines) glColor4f( 1., 1., 1., 1. );
		glBegin( GL_TRIANGLE_STRIP );
		for (s=c->firstSegment; s!=NULL; s=s->next)
		{
			i=s->i, j=s->j, k=s->k, last=s->last;
			switch (s->which)
			{
				case LEFT:
					DRAWIT( i, j, k, i+1, j, k, last-k+1, k++ );
					break;
		
				case RIGHT:
					DRAWIT( i, j+1, k+1, i+1, j+1, k+1, k-last+1, k-- );
					break;
		
				case TOP:
					DRAWIT( i, j, k+1, i+1, j, k+1, last-j+1, j++ );
					break;
		
				case BOTTOM:
					DRAWIT( i, j+1, k, i+1, j+1, k, j-last+1, j-- );
					break;
		
				case FRONT:
					DRAWIT_CLOSED( i+1, j, k, i+1, j+1, k, last-k+1, k++ );
					break;
		
				case BACK:
					DRAWIT_CLOSED( i, j+1, k, i, j, k, last-k+1, k++ );
					break;
			}
		}
/*
		close the contour
		-----------------
*/
		s = c->firstSegment;
		if (s->which == LEFT)
		{
			DRAWIT( s->i, s->j, s->k, s->i+1, s->j, s->k, 1, k++ );
		}
		glEnd();
	}
	glDisable( GL_TEXTURE_2D );
}

/*---------------------------------------------------------------------------*/
/*	drawOutlines
/*---------------------------------------------------------------------------*/
void
drawOutlines( void )
{
	int count, ndx;
	Segment *s;
	Contour *c;

#ifdef GL_EXT_polygon_offset
	if (polygonoffset)
	{
    	glEnable(GL_POLYGON_OFFSET_EXT);
	    glPolygonOffsetEXT(  -.5, -.03*nearclip/farclip );
	}
#endif
	if (antialiased)
	{
		glEnable( GL_LINE_SMOOTH );
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glHint( GL_LINE_SMOOTH_HINT, GL_NICEST );
		glLineWidth( 3. );
	}
	glDisable( GL_STENCIL_TEST );

	/*=================================================================*/
	if (polygonmode) {
	/*=================================================================*/

	glPolygonMode( GL_FRONT_AND_BACK, GL_LINE );

#undef DRAWIT
#define DRAWIT( i0, j0, k0, i1, j1, k1, num, change )\
    count = num;\
    while (count--)\
    {\
        glVertex3i( i0, j0, k0 );\
        glVertex3i( i1, j1, k1 );\
        change;\
    }

#undef DRAWIT_CLOSED
#define DRAWIT_CLOSED( i0, j0, k0, i1, j1, k1, num, change )\
	DRAWIT( i0, j0, k0, i1, j1, k1, num, change )\
	glVertex3i( i0, j0, k );\
	glVertex3i( i1, j1, k );

	for (c=firstContour; c!=NULL; c=c->next)
	{
		int i, j, k, last;
		if (hiddenlines) glColor3f( 1., 1., 1. );
		else glColor3fv( &RGBA[0] );

		glBegin( GL_QUAD_STRIP );
		for (s=c->firstSegment; s!=NULL; s=s->next)
		{
			i=s->i, j=s->j, k=s->k, last=s->last;
			switch (s->which)
			{
				case LEFT:
					DRAWIT( i, j, k, i+1, j, k, last-k+1, k++ );
					break;
		
				case RIGHT:
					DRAWIT( i, j+1, k+1, i+1, j+1, k+1, k-last+1, k-- );
					break;
		
				case TOP:
					DRAWIT( i, j, k+1, i+1, j, k+1, last-j+1, j++ );
					break;
		
				case BOTTOM:
					DRAWIT( i, j+1, k, i+1, j+1, k, j-last+1, j-- );
					break;
		
				case FRONT:
					DRAWIT_CLOSED( i+1, j, k, i+1, j+1, k, last-k+1, k++ );
					break;
		
				case BACK:
					DRAWIT_CLOSED( i, j+1, k, i, j, k, last-k+1, k++ );
					break;
			}
		}
/*
		close the contour
		-----------------
*/
		s = c->firstSegment;
		if (s->which == LEFT)
		{
			DRAWIT( s->i, s->j, s->k, s->i+1, s->j, s->k, 1, k++ );
		}
		glEnd();
	}

	glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );

	/*=================================================================*/
	} else {	/* NO polygonmode */
	/*=================================================================*/

#undef DRAWIT
#define DRAWIT( i0, j0, k0, i1, j1, k1, num, change )\
    count = num;\
    while (count--)\
    {\
        glVertex3i( i0, j0, k0 );\
        glVertex3i( i1, j1, k1 );\
        change;\
    }

#undef DRAWIT_CLOSED
#define DRAWIT_CLOSED( i0, j0, k0, i1, j1, k1, num, change )\
	DRAWIT( i0, j0, k0, i1, j1, k1, num, change )\
	glVertex3i( i0, j0, k );\
	glVertex3i( i1, j1, k );

	for (c=firstContour; c!=NULL; c=c->next)
	{
		if (hiddenlines) glColor3f( 1., 1., 1. );
		else glColor3fv( &RGBA[0] );

		glBegin( GL_LINES );
		for (s=c->firstSegment; s!=NULL; s=s->next)
		{
			int i=s->i, j=s->j, k=s->k, last=s->last;
			switch (s->which)
			{
				case LEFT:
					glVertex3i( i, j, k );
					glVertex3i( i, j, last+1 );
					glVertex3i( i+1, j, k );
					glVertex3i( i+1, j, last+1 );
					DRAWIT( i, j, k, i+1, j, k, last-k+1, k++ );
					break;
		
				case RIGHT:
					glVertex3i( i, j+1, last );
					glVertex3i( i, j+1, k+1 );
					glVertex3i( i+1, j+1, last );
					glVertex3i( i+1, j+1, k+1 );
					DRAWIT( i, j+1, k+1, i+1, j+1, k+1, k-last+1, k-- );
					break;
		
				case TOP:
					glVertex3i( i, j, k+1 );
					glVertex3i( i, last+1, k+1 );
					glVertex3i( i+1, j, k+1 );
					glVertex3i( i+1, last+1, k+1 );
					DRAWIT( i, j, k+1, i+1, j, k+1, last-j+1, j++ );
					break;
		
				case BOTTOM:
					glVertex3i( i, j+1, k );
					glVertex3i( i, last, k );
					glVertex3i( i+1, j+1, k );
					glVertex3i( i+1, last, k );
					DRAWIT( i, j+1, k, i+1, j+1, k, j-last+1, j-- );
					break;
		
				case FRONT:
					glVertex3i( i+1, j, k );
					glVertex3i( i+1, j, last+1 );
					glVertex3i( i+1, j+1, k );
					glVertex3i( i+1, j+1, last+1 );
					DRAWIT_CLOSED( i+1, j, k, i+1, j+1, k, last-k+1, k++ );
					break;
		
				case BACK:
					glVertex3i( i, j+1, k );
					glVertex3i( i, j+1, last+1 );
					glVertex3i( i, j, k );
					glVertex3i( i, j, last+1 );
					DRAWIT_CLOSED( i, j+1, k, i, j, k, last-k+1, k++ );
					break;
			}
		}
		glEnd();
	}
	/*=================================================================*/
	}	/* polygonmode */
	/*=================================================================*/

#ifdef GL_EXT_polygon_offset
    glDisable(GL_POLYGON_OFFSET_EXT);
#endif
	glDisable( GL_LINE_SMOOTH );
	glBlendFunc( GL_ONE, GL_ZERO );
	glDisable(GL_BLEND);
}

/*---------------------------------------------------------------------------*/
/*	drawCappedPlane
/*---------------------------------------------------------------------------*/
void
drawCappedPlane( void )
{
	float lb[3], lt[3], rb[3], rt[3], scale;
	static int first = GL_TRUE;

	updatePlaneStructure( &clipplane );

	scale = -Plane.Equation[3]*maxdim;
    clipplane.farCentre[0] = centre.x + scale*clipplane.Equation[0];
    clipplane.farCentre[1] = centre.y + scale*clipplane.Equation[1];
    clipplane.farCentre[2] = centre.z + scale*clipplane.Equation[2];

	SCLVEC3( clipplane.up, maxdim );
	SCLVEC3( clipplane.dn, maxdim );

#define SETCORNER( v, corner )\
	v[0]=corner(clipplane,0);\
	v[1]=corner(clipplane,1);\
	v[2]=corner(clipplane,2);

	SETCORNER( lb, LB );
	SETCORNER( lt, LT );
	SETCORNER( rb, RB );
	SETCORNER( rt, RT );

	glDisable( GL_CLIP_PLANE0 );
	glDisable( GL_DEPTH_TEST );
	glEnable( GL_STENCIL_TEST );
	glStencilFunc( GL_EQUAL, 1, 1 );
	glStencilOp( GL_KEEP, GL_KEEP, GL_KEEP );

	glEnable( GL_CULL_FACE );
	glCullFace( GL_BACK );

	if (!hiddenlines)
	{
		bindTexture( GLOBAL_TEXTURE );
		glColor3f( 1., 1., 1. );
	}
	else
	{
		bindTexture( OUTLINE_3D_TEXTURE );
		glColor3f( 0., 0., 1. );
	}

	glBegin( GL_TRIANGLE_STRIP );
		glTexCoord3fv(lb); glVertex3fv( lb );
		glTexCoord3fv(lt); glVertex3fv( lt );
		glTexCoord3fv(rb); glVertex3fv( rb );
		glTexCoord3fv(rt); glVertex3fv( rt );
	glEnd();

	glDisable( GL_TEXTURE_COLOR_TABLE_SGI );
	glDisable( GL_TEXTURE_3D_EXT );
	glDisable( GL_STENCIL_TEST );
	glColor3f( 0., 0., 1. );
	glPolygonMode( GL_FRONT_AND_BACK, GL_LINE );
	glBegin( GL_QUADS );
		glVertex3fv( lb );
		glVertex3fv( lt );
		glVertex3fv( rt );
		glVertex3fv( rb );
	glEnd();
	glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );

	glDisable( GL_TEXTURE_COLOR_TABLE_SGI );
	glDisable( GL_TEXTURE_3D_EXT );
	glDisable( GL_STENCIL_TEST );
	glDisable( GL_CULL_FACE );
	glEnable( GL_DEPTH_TEST );

}

/*---------------------------------------------------------------------------*/
/*	setCappedPlane
/*---------------------------------------------------------------------------*/
void
setCappedPlane( void )
{
	GLdouble	x, y, z, offset;

	clipplane.Equation[0] = Plane.Equation[0];
	clipplane.Equation[1] = Plane.Equation[1];
	clipplane.Equation[2] = Plane.Equation[2];
	offset = -Plane.Equation[3]*maxdim;

	x = centre.x + offset*Plane.Equation[0];
	y = centre.y + offset*Plane.Equation[1];
	z = centre.z + offset*Plane.Equation[2];

	clipplane.Equation[3] =
	-(x*clipplane.Equation[0] + y*clipplane.Equation[1] + z*clipplane.Equation[2]);

	glPushMatrix();
		glLoadIdentity();
		glTranslatef( -obs.x, -obs.y, -obs.z );
		glMultMatrixf( rotationMatrix );
		glTranslatef( -centre.x, -centre.y, -centre.z );
		glClipPlane( GL_CLIP_PLANE0, clipplane.Equation );
		glEnable( GL_CLIP_PLANE0 );
	glPopMatrix();

	glClearStencil( 0 );
	glClear( GL_STENCIL_BUFFER_BIT );
	glEnable( GL_STENCIL_TEST );
	glStencilFunc( GL_ALWAYS, 0, 1 );
	glStencilOp( GL_INVERT, GL_INVERT, GL_INVERT );
}

/*---------------------------------------------------------------------------*/
/*	makePopupMenu
/*---------------------------------------------------------------------------*/
#define	OUT_LINE_ITEM	1
#define	OUT_POLY_ITEM	2
#define	OUT_TEX_ITEM	3
#define	OUT_OFF_ITEM	4

#define	TEX_LINEAR_ITEM		11
#define	TEX_NEAREST_ITEM	12
#define	TEX_OFF_ITEM		13

#define MIPMAP_LINEAR_ITEM	21
#define	MIPMAP_NEAREST_ITEM 22
#define	MIPMAP_OFF_ITEM		23

#define ANTIALIAS_ON_ITEM	31
#define ANTIALIAS_OFF_ITEM	32

#define HIDDENLINE_ON_ITEM	41
#define HIDDENLINE_OFF_ITEM	42

#define CAPPING_ON_ITEM		51
#define CAPPING_OFF_ITEM	52

#define OFFSET_ON_ITEM		61
#define OFFSET_OFF_ITEM		62

makePopupMenu()
{
	int outlining_menu,
		texture_menu,
		mipmap_menu,
		antialias_menu,
		hiddenline_menu,
		capping_menu,
		offset_menu,
		MAIN_menu;

	outlining_menu = glutCreateMenu( menuCallBack );
	glutAddMenuEntry( "Lines->Polygons", OUT_POLY_ITEM );

	texture_menu = glutCreateMenu( menuCallBack );
	glutAddMenuEntry( "Nearest->Linear", TEX_LINEAR_ITEM );

	mipmap_menu = glutCreateMenu( menuCallBack );
	glutAddMenuEntry( "Linear->Off", MIPMAP_OFF_ITEM );

	antialias_menu = glutCreateMenu( menuCallBack );
	glutAddMenuEntry( "On->Off", ANTIALIAS_OFF_ITEM );

	hiddenline_menu = glutCreateMenu( menuCallBack );
	glutAddMenuEntry( "Off->On", HIDDENLINE_ON_ITEM );

	capping_menu = glutCreateMenu( menuCallBack );
	glutAddMenuEntry( "On->Off", CAPPING_OFF_ITEM );

	offset_menu = glutCreateMenu( menuCallBack );
	glutAddMenuEntry( "On->Off", OFFSET_OFF_ITEM );

	MAIN_menu = glutCreateMenu( menuCallBack );
	glutAddSubMenu( "Anti-Aliasing", antialias_menu );
	glutAddSubMenu( "Capping", capping_menu );
	glutAddSubMenu( "Hidden-Lines", hiddenline_menu );
	glutAddSubMenu( "Polygon-Offset", offset_menu );
	glutAddSubMenu( "OutLining", outlining_menu );
	glutAddSubMenu( "Texturing", texture_menu );
	glutAddSubMenu( "MipMaping", mipmap_menu );

	glutAttachMenu( GLUT_RIGHT_BUTTON );
}

void
menuCallBack( int item )
{
	switch (item)
	{
		case OUT_LINE_ITEM:
			glutChangeToMenuEntry( 1, "Lines->Polygons", OUT_POLY_ITEM );
			outlined = GL_TRUE;
			polygonmode = GL_FALSE;
			outtextured = GL_FALSE;
			break;
		
		case OUT_POLY_ITEM:
			glutChangeToMenuEntry( 1, "Polygons->Texture", OUT_TEX_ITEM );
			outlined = GL_TRUE;
			polygonmode = GL_TRUE;
			outtextured = GL_FALSE;
			break;
		
		case OUT_TEX_ITEM:
			glutChangeToMenuEntry( 1, "Texture->Off", OUT_OFF_ITEM );
			outlined = GL_FALSE;
			polygonmode = GL_FALSE;
			outtextured = GL_TRUE;
			break;
		
		case OUT_OFF_ITEM:
			glutChangeToMenuEntry( 1, "Off->Lines", OUT_LINE_ITEM );
			outlined = GL_FALSE;
			polygonmode = GL_FALSE;
			outtextured = GL_FALSE;
			break;
		
		case TEX_NEAREST_ITEM:
			glutChangeToMenuEntry( 1, "Nearest->Linear", TEX_LINEAR_ITEM );
			textureMode = GL_NEAREST;
			switch (minfilter)
			{
				case GL_NEAREST: break;
				case GL_LINEAR: minfilter = GL_NEAREST; break;
				case GL_NEAREST_MIPMAP_NEAREST: break;
				case GL_NEAREST_MIPMAP_LINEAR: break;
				case GL_LINEAR_MIPMAP_NEAREST: minfilter = GL_NEAREST_MIPMAP_NEAREST; break;
				case GL_LINEAR_MIPMAP_LINEAR: minfilter = GL_NEAREST_MIPMAP_LINEAR; break;
			}
			break;
		
		case TEX_LINEAR_ITEM:
			glutChangeToMenuEntry( 1, "Linear->Off", TEX_OFF_ITEM );
			textureMode = GL_LINEAR;
			switch (minfilter)
			{
				case GL_NEAREST: minfilter = GL_LINEAR; break;
				case GL_LINEAR: break;
				case GL_NEAREST_MIPMAP_NEAREST: minfilter = GL_LINEAR_MIPMAP_NEAREST; break;
				case GL_NEAREST_MIPMAP_LINEAR: minfilter = GL_LINEAR_MIPMAP_LINEAR; break;
				case GL_LINEAR_MIPMAP_NEAREST: break;
				case GL_LINEAR_MIPMAP_LINEAR: break;
			}
			break;

		case TEX_OFF_ITEM:
			glutChangeToMenuEntry( 1, "Off->Nearest", TEX_NEAREST_ITEM );
			textureMode = GL_FALSE;
			break;
		
		case MIPMAP_NEAREST_ITEM:
			glutChangeToMenuEntry( 1, "Nearest->Linear", MIPMAP_LINEAR_ITEM );
			switch (minfilter)
			{
				case GL_NEAREST: minfilter = GL_NEAREST_MIPMAP_NEAREST; break;
				case GL_LINEAR: minfilter = GL_LINEAR_MIPMAP_NEAREST; break;
				case GL_NEAREST_MIPMAP_NEAREST: break;
				case GL_NEAREST_MIPMAP_LINEAR: minfilter = GL_NEAREST_MIPMAP_NEAREST; break;
				case GL_LINEAR_MIPMAP_NEAREST: break;
				case GL_LINEAR_MIPMAP_LINEAR: minfilter = GL_LINEAR_MIPMAP_NEAREST; break;
			}
			break;

		case MIPMAP_LINEAR_ITEM:
			glutChangeToMenuEntry( 1, "Linear->Off", MIPMAP_OFF_ITEM );
			switch (minfilter)
			{
				case GL_NEAREST: minfilter = GL_NEAREST_MIPMAP_LINEAR; break;
				case GL_LINEAR: minfilter = GL_LINEAR_MIPMAP_LINEAR; break;
				case GL_NEAREST_MIPMAP_NEAREST: minfilter = GL_NEAREST_MIPMAP_LINEAR; break;
				case GL_NEAREST_MIPMAP_LINEAR: break;
				case GL_LINEAR_MIPMAP_NEAREST: minfilter = GL_LINEAR_MIPMAP_LINEAR; break;
				case GL_LINEAR_MIPMAP_LINEAR: break;
			}
			break;

		case MIPMAP_OFF_ITEM:
			glutChangeToMenuEntry( 1, "Off->Nearest", MIPMAP_NEAREST_ITEM );
			switch (minfilter)
			{
				case GL_NEAREST: break;
				case GL_LINEAR: break;
				case GL_NEAREST_MIPMAP_NEAREST: minfilter = GL_NEAREST; break;
				case GL_NEAREST_MIPMAP_LINEAR: minfilter = GL_NEAREST; break;
				case GL_LINEAR_MIPMAP_NEAREST: minfilter = GL_LINEAR; break;
				case GL_LINEAR_MIPMAP_LINEAR: minfilter = GL_LINEAR; break;
			}
			break;
		
		case ANTIALIAS_ON_ITEM:
			glutChangeToMenuEntry( 1, "On->Off", ANTIALIAS_OFF_ITEM );
			antialiased = GL_TRUE;
			break;

		case ANTIALIAS_OFF_ITEM:
			glutChangeToMenuEntry( 1, "Off->On", ANTIALIAS_ON_ITEM );
			antialiased = GL_FALSE;
			break;
		
		case HIDDENLINE_ON_ITEM:
			glutChangeToMenuEntry( 1, "On->Off", HIDDENLINE_OFF_ITEM );
			hiddenlines = GL_TRUE;
			break;

		case HIDDENLINE_OFF_ITEM:
			glutChangeToMenuEntry( 1, "Off->On", HIDDENLINE_ON_ITEM );
			hiddenlines = GL_FALSE;
			break;
		
		case CAPPING_ON_ITEM:
			glutChangeToMenuEntry( 1, "On->Off", CAPPING_OFF_ITEM );
			capping = GL_TRUE;
			break;
			
		case CAPPING_OFF_ITEM:
			glutChangeToMenuEntry( 1, "Off->On", CAPPING_ON_ITEM );
			capping = GL_FALSE;
			break;
		
		case OFFSET_ON_ITEM:
			glutChangeToMenuEntry( 1, "On->Off", OFFSET_OFF_ITEM );
			polygonoffset = GL_TRUE;
			break;

		case OFFSET_OFF_ITEM:
			glutChangeToMenuEntry( 1, "Off->On", OFFSET_ON_ITEM );
			polygonoffset = GL_FALSE;
			break;
	}
	redraw = 1;
}
/*---------------------------------------------------------------------------*/
/*	bindTexture
/*---------------------------------------------------------------------------*/
static void
bindTexture( int which )
{
	static int first = 1;
	static GLuint texobj[2];

	if (!textureMode) return;

/*
	Texture Definition
	------------------
*/
	if (first)
	{
		int i, j, k;
		static u_char outline3D[TEXX*TEXY*TEXZ];
		static u_char *ptr, outline2D[TEXX*TEXY];

		glGenTexturesEXT( 3, texobj );

		for (i=0; i<TEXX; i++)
		for (j=0; j<TEXY; j++)
		{
			outline2D[TEXY*j+i] = (i%(TEXX-1)) && (j%(TEXY-1)) ? 0 : 0xff;
		}

		glBindTextureEXT( GL_TEXTURE_2D, texobj[0] );
		gluBuild2DMipmaps(
			GL_TEXTURE_2D,		/* GLenum target, */
			1,					/* GLint component, */
			TEXX,				/* GLsizei width, */
			TEXY,				/* GLsizei height, */
			GL_LUMINANCE,		/* GLenum format, */
			GL_UNSIGNED_BYTE,	/* GLenum type, */
			outline2D			/* const GLvoid *pixels */
		);

		for (i=0; i<TEXX; i++)
		for (j=0; j<TEXY; j++)
		for (k=0; k<TEXZ; k++)
		{
			int I=i%(TEXX-1), J=j%(TEXY-1), K=k%(TEXZ-1);
			outline3D[TEXX*(TEXY*k+j) + i] = (I && J && K) ? 0 : 0xff;
		}

		glBindTextureEXT( GL_TEXTURE_3D_EXT, texobj[1] );
		glTexImage3DEXT(
			GL_TEXTURE_3D_EXT,	/* GLenum target, */
			0,					/* GLint level, */
			GL_LUMINANCE8_EXT,	/* GLenum internalformat, */
			TEXX,				/* GLint width, */
			TEXY,				/* GLsizei height, */
			TEXZ,				/* GLsizei depth, */
			0,					/* GLint border, */
			GL_LUMINANCE,		/* GLenum format, */
			GL_UNSIGNED_BYTE,	/* GLenum type, */
			outline3D			/* const GLvoid *pixels */
		);
		
		glBindTextureEXT( GL_TEXTURE_3D_EXT, texobj[2] );
		glTexImage3DEXT(
			GL_TEXTURE_3D_EXT,	/* GLenum target, */
			0,					/* GLint level, */
			GL_RGB,				/* GLenum internalformat, */
			dimx,				/* GLint width, */
			dimy,				/* GLsizei height, */
			dimz,				/* GLsizei depth, */
			0,					/* GLint border, */
			GL_LUMINANCE,		/* GLenum format, */
			GL_UNSIGNED_BYTE,	/* GLenum type, */
			scalar				/* const GLvoid *pixels */
		);

		first = GL_FALSE;
	}

/*
	Texture Binding
	---------------
*/
	switch (which)
	{

	case OUTLINE_2D_TEXTURE:
	/**********************/

		glBindTextureEXT( GL_TEXTURE_2D, texobj[0] );
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, textureMode);
		glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, minfilter );

		glMatrixMode( GL_TEXTURE );
		glLoadIdentity();
		glMatrixMode( GL_MODELVIEW );

		if (hiddenlines)
		{
			/* GL_REPLACE_EXT NOT implemented on IMPACT */
			glTexEnvf ( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );
		}
		else
		{
			glTexEnvf ( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_BLEND );
			glTexEnvfv( GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, &RGBA[0] );
		}
		glEnable( GL_TEXTURE_2D );
		break;

	case OUTLINE_3D_TEXTURE:
	/**********************/

		glBindTextureEXT( GL_TEXTURE_3D_EXT, texobj[1] );

		glTexParameterf(GL_TEXTURE_3D_EXT, GL_TEXTURE_WRAP_S, GL_REPEAT);
		glTexParameterf(GL_TEXTURE_3D_EXT, GL_TEXTURE_WRAP_T, GL_REPEAT);
		glTexParameterf(GL_TEXTURE_3D_EXT, GL_TEXTURE_WRAP_R_EXT,GL_REPEAT);
		glTexParameterf(GL_TEXTURE_3D_EXT, GL_TEXTURE_MAG_FILTER, textureMode);
		glTexParameterf( GL_TEXTURE_3D_EXT, GL_TEXTURE_MIN_FILTER, textureMode);

		glTexEnvf ( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );

		glMatrixMode( GL_TEXTURE );
		glLoadIdentity();
		glMatrixMode( GL_MODELVIEW );

		glEnable( GL_TEXTURE_3D_EXT );
		break;

	case GLOBAL_TEXTURE:
	/******************/

		if (hiddenlines) return;

		glBindTextureEXT( GL_TEXTURE_3D_EXT, texobj[2] );
		glTexParameteri(GL_TEXTURE_3D_EXT, GL_TEXTURE_WRAP_S, GL_CLAMP);
		glTexParameteri(GL_TEXTURE_3D_EXT, GL_TEXTURE_WRAP_T, GL_CLAMP);
		glTexParameteri(GL_TEXTURE_3D_EXT, GL_TEXTURE_WRAP_R_EXT, GL_CLAMP);
		glTexParameterf(GL_TEXTURE_3D_EXT, GL_TEXTURE_MAG_FILTER, textureMode);
		glTexParameterf(GL_TEXTURE_3D_EXT, GL_TEXTURE_MIN_FILTER, textureMode);

		glTexEnvf ( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );
		/* GL_REPLACE_EXT NOT implemented on IMPACT */

		glMatrixMode( GL_TEXTURE );
		glLoadIdentity();
		glScalef( 1./dimx, 1./dimy, 1./dimz );
		glMatrixMode( GL_MODELVIEW );

		glEnable( GL_TEXTURE_3D_EXT );

		if (updateTLUT) loadTLUT();
		glEnable( GL_TEXTURE_COLOR_TABLE_SGI );
		break;
	}
}

/*---------------------------------------------------------------------------*/
/*	loadTLUT
/*---------------------------------------------------------------------------*/
static void
loadTLUT( void )
{
	int i, num;
	static float colortable[256][3];
	int curwin = glutGetWindow();

	num = (RGBA == (float *)Default_RGBA) ? 14 : 256;

	for (i=0; i<num; i++)
	{
		colortable[i][0] = RGBA[4*i+0];
		colortable[i][1] = RGBA[4*i+1];
		colortable[i][2] = RGBA[4*i+2];
	}

	glutSetWindow( GISwin );
	glColorTableSGI(
		GL_TEXTURE_COLOR_TABLE_SGI,
		GL_RGB,
		256,
		GL_RGB,
		GL_FLOAT,
		colortable
	);
	glutSetWindow( curwin );
	updateTLUT = GL_FALSE;
}
