 /**************************************************************************
  *                                                                        *
  *               Copyright (C) 1996, Silicon Graphics, Inc.               *
  *                          All Rights Reserved.                          *
  *                                                                        *
  *  The files in this subtree  contain  UNPUBLISHED  PROPRIETARY  SOURCE  *
  *  CODE of Silicon Graphics, Inc.;  the  contents  of  these  files may  *
  *  not be disclosed to third  parties,  copied  or  duplicated  in  any  *
  *  form, in whole or in part,  without  the  prior  written  permission  * 
  *  of  Silicon Graphics, Inc.                                            *
  *                                                                        *
  *  RESTRICTED RIGHTS LEGEND:                                             *
  *  Use,  duplication  or  disclosure  by  the  Government is subject to  *
  *  restrictions as set forth in subdivision (c)(1)(ii) of the Rights in  *
  *  Technical Data and Computer Software clause at  DFARS  252.227-7013,  *
  *  and/or in similar or successor  clauses in the FAR,  DOD or NASA FAR  *
  *  Supplement.  Unpublished - rights reserved  under the Copyright Laws  *
  *  of the United States.                                                 *
  *                                                                        *
  *  THIS SOFTWARE IS  PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,  *
  *  EXPRESS,  IMPLIED OR  OTHERWISE,  INCLUDING  WITHOUT LIMITATION, ANY  *
  *  WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.      *
  *                                                                        *
  *  IN  NO  EVENT  SHALL  SILICON  GRAPHICS  BE  LIABLE FOR ANY SPECIAL,  *
  *  INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF  ANY  KIND,  OR ANY  *
  *  DAMAGES WHATSOEVER RESULTING FROM LOSS  OF  USE,  DATA  OR  PROFITS,  *
  *  WHETHER OR NOT ADVISED OF  THE  POSSIBILITY  OF  DAMAGE,  AND ON ANY  *
  *  THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR  *
  *  PERFORMANCE OF THIS SOFTWARE.                                         *
  **************************************************************************/

/*
    Author : Patrick Bouchaud
             galaad@neu.sgi.com
*/

/*===========================================================================*/
/*
/*	Extract.c :
/*
/*	DONNEES EN ENTREE:
/*	------------------
/*	La routine ExtractContour s'interface avec un bloc de donnees
/*	de dimensions imax, jmax, kmax
/*	au travers des routines (*testFunc)()
/*
/*	Cette implementation utilise l'include Extract.h ou
/*	imax jmax et kmax sont des references externes d'entiers
/*	et les routines d'interface sont des references externes de fonctions
/*
/*	DONNEES EN SORTIE:
/*	------------------
/*	La routine ExtractContour renvoie un pointeur sur une liste chainee de
/*	donnees graphiques agencees par SetGeom
/*	Ces donnees correspondent a des bandes (qstrip) dans les plans (i=cste)
/*	et (j=cste).
/*	Ces bandes definissent les contours des blocs de cellules valides
/*	(validitee obtenue en retour de (*testFunc)) appartenant a ces plans.
/*	Les bandes des plans (i=cste) correspondent a des contours fermes.
/*
/*===========================================================================*/

#include	<stdio.h>
#include	<sys/types.h>
#include	<malloc.h>

#include	"extractor.h"
#include	"mallocbis.h"

#define I_AXIS          1
#define J_AXIS          2
#define K_AXIS          3
#define KBIS_AXIS       4

typedef struct ListType {
	int             axis,
			firstj,
			firstk,
			dernier_ndx;
        struct ListType *next,
			*last;
} ListStruct;

static ListStruct *GetList_GchInvalide( int i, int j );
static int	GetSeg_GchInvalide( int i, int j, int k );
static int	GetSeg_DteInvalide( int i, int j, int k );
static int	GetSeg_SusInvalide( int i, int j, int k );
static int	GetSeg_SousInvalide( int i, int j, int k );
static int	GetSeg_DerInvalide( int i, int j, int k );
static int	GetSeg_DevInvalide( int i, int j, int k );

typedef struct FrontierStruct
{
	int ndx;
	struct FrontierStruct *next, *last;

} Frontier;
static Frontier	**frontier = NULL,
				**last_frontier = NULL;

static ListStruct	**Seg_GchInvalide = NULL;

static int	ifirst, jfirst, kfirst,
			imax, jmax, kmax;
static void	(*bgnFunc)(void),
			(*endFunc)(void),
			(*setFunc)(int,int,int,int,int);
static int	(*testFunc)(int,int,int);

ListStruct *freelistptr = NULL;
Frontier *freefrptr = NULL;

#define	newFrontier(fr) newItem( fr, freefrptr, Frontier )
#define	freeFrontier(fr) freeItem( fr, freefrptr )
#define	newList(liste) newItem( liste, freelistptr, ListStruct )
#define	freeList(liste) freeItem( liste, freelistptr )

/*---------------------------------------------------------------------------*/
/*				Extract
/*---------------------------------------------------------------------------*/
ExtractContour( Extractor *extractor )
{
 int	i, j, jnum = extractor->jnum;

 ifirst = extractor->ifirst;
 jfirst = extractor->jfirst;
 kfirst = extractor->kfirst;
 imax	= ifirst + extractor->inum;
 jmax	= jfirst + extractor->jnum;
 kmax	= kfirst + extractor->knum;
 bgnFunc = extractor->bgnFunc;
 endFunc = extractor->endFunc;
 setFunc = extractor->setFunc;
 testFunc = extractor->testFunc;

 if ( Seg_GchInvalide != NULL )
	free( Seg_GchInvalide );
 Seg_GchInvalide = (ListStruct **) calloc(jnum, sizeof(ListStruct *));

 if (frontier != NULL)
 {
 	free( frontier );
 	free( last_frontier );
 }
 frontier = (Frontier **) calloc( jnum, sizeof(Frontier *) );
 last_frontier = (Frontier **) calloc( jnum, sizeof(Frontier *) );

/*
 On travaille dans le repere suivant :
 -------------------------------------

		K
		^
		|
		|  
		| /
		|/
		x--------------------> J
       /
      /
     /
    L
   I
*/

 for ( i=ifirst; i<imax; i++ )
 {
/*
	Pour chaque plan (I=cste) Creation de la liste des segments de cellules
	verticaux bordes a gauche de cellules invalides
*/
	for ( j=jfirst; j<jmax; j++ ) {
	 	Seg_GchInvalide[j] = GetList_GchInvalide( i, j );
	}
/*
	on complemente ces listes de segments de manieres a en faire des
	lignes FERMEES.
	On convertit ces listes au format GRAPHIQUE.
	( c.a.d. on complete chaque liste avec les cellules horizontales et
	les cellules verticales bordees a droite de cellules invalides dans
	ce plan )
*/
	Complete_List( i );
/*
	Creation des listes de segments de cellules verticaux bordes devant
	ou derriere de cellules invalides
*/
#ifdef OPT
	BuildPatches( i );
#endif
 }
 BuildPatches( imax );

} /* Extract */

/*---------------------------------------------------------------------------*/
/*			GetList_GchInvalide
/*---------------------------------------------------------------------------*/
static ListStruct *
GetList_GchInvalide( int i, int j )
{
 int	k,
	liste_en_cours = 0;
 ListStruct
	Root,
	*liste_courante = &Root;

 Root.next = Root.last = NULL;

 for ( k=kfirst; k<kmax; k++ ) {

	if ( (*testFunc)(i,j,k) && !(*testFunc)(i,j-1,k) ) {
		if ( liste_en_cours ) {
			liste_courante->dernier_ndx = k;
		} else {
			newList( liste_courante->next );
			liste_courante->next->last = liste_courante;
			liste_courante = liste_courante->next;

			liste_courante->next = NULL;
			liste_courante->axis = K_AXIS;
			liste_courante->firstj = j;
			liste_courante->firstk =
			liste_courante->dernier_ndx = k;
			liste_en_cours = 1;
		}
	} else {
		liste_en_cours = 0;
	}
 }

 if ( Root.next != NULL ) {
	Root.next->last = NULL;
	return Root.next;
 } else {
	return NULL;
 }
} /* GetList_GchInvalide */

/*---------------------------------------------------------------------------*/
/*			insert_frontier
/*---------------------------------------------------------------------------*/
insert_frontier( int j, int k )
{
	Frontier *fr = frontier[j];
	if (fr == NULL)
	{
		newFrontier( fr );
		fr->next = NULL;
		fr->last = NULL;
		fr->ndx = k;
		frontier[j] = fr;
	}
	else if (fr->ndx > k)
	{
		Frontier *new_fr;
		newFrontier( new_fr );
		new_fr->next = fr;
		new_fr->last = NULL;
		new_fr->ndx = k;
		fr->last = new_fr;
		frontier[j] = new_fr;
	}
	else
	{
		while ((fr->next==NULL) ? 0 : (fr->ndx < k))
			fr = fr->next;
		if (fr->ndx > k)
		{
			Frontier *new_fr;
			newFrontier( new_fr );
			new_fr->next = fr;
			new_fr->last = fr->last;
			new_fr->ndx = k;
			fr->last->next = new_fr;
			fr->last = new_fr;
		}
		else
		{
			Frontier *new_fr;
			newFrontier( new_fr );
			new_fr->last = fr;
			new_fr->next = fr->next;
			new_fr->ndx = k;
			if (fr->next != NULL)
				fr->next->last = new_fr;
			fr->next = new_fr;
		}
	}
}

/*---------------------------------------------------------------------------*/
/*			swap_frontiers
/*---------------------------------------------------------------------------*/
swap_frontiers()
{
	int j;
	Frontier	**save_frontier = last_frontier;
	for (j=0; j<jmax; j++)
	{
		Frontier *fr = last_frontier[j];
		if ( fr != NULL )
		{
			while (fr->next != NULL)
			{
				fr = fr->next;
				freeFrontier( fr->last );
			}
			freeFrontier( fr );
		}
		last_frontier[j] = NULL;
	}
	last_frontier = frontier;
	frontier = save_frontier;
}

/*---------------------------------------------------------------------------*/
/*			BuildPatches
/*---------------------------------------------------------------------------*/
#ifdef OPT
BuildPatches( int i )
{
	register int ilast = i-1, j;

/*
	first data plane
	----------------
*/
	if (i==0)
	{
		for (j=0; j<jmax; j++)
		{
			Frontier *fr = frontier[j];
			while (fr != NULL)
			{
				(*bgnFunc)();
				(*setFunc)( BACK, i, j, fr->ndx, fr->next->ndx );
				(*endFunc)();
				fr = fr->next->next;
			}
		}
		swap_frontiers();
		return;
	}
/*
	last data plane
	---------------
*/
	if (i==imax)
	{
		for (j=0; j<jmax; j++)
		{
			Frontier *fr = last_frontier[j];
			while (fr != NULL)
			{
				(*bgnFunc)();
				(*setFunc)( FRONT, ilast, j, fr->ndx, fr->next->ndx );
				(*endFunc)();
				fr = fr->next->next;
			}
		}
		swap_frontiers();
		return;
	}

/*
	intermediate data planes
	------------------------
*/
#define	SETFRONT( a, b )\
	(*bgnFunc)(); (*setFunc)(FRONT,i-1,j,a,b); (*endFunc)();

#define	SETBACK( a, b )\
	(*bgnFunc)(); (*setFunc)(BACK,i,j,a,b); (*endFunc)();

#define	NEXT( f, a, b )\
	if ((f=f->next->next) != NULL)\
	{\
		a = f->ndx;\
		b = f->next->ndx;\
	}

	for (j=0; j<jmax; j++)
	{
		Frontier *fr, *cfr;
		int start, end, clip_start, clip_end;
/*
		First draw all the FRONT segments of last plane
		-----------------------------------------------
*/
		fr = last_frontier[j];
		cfr = frontier[j];
		if (fr != NULL)
		{
			start = fr->ndx;
			end = fr->next->ndx;
		}
		if (cfr != NULL)
		{
			clip_start = cfr->ndx;
			clip_end = cfr->next->ndx;
		}

		while ((fr != NULL) && (cfr != NULL))
		{
			while ((fr==NULL) ? 0 : (end < clip_start))
			{
				SETFRONT( start, end );
				NEXT( fr, start, end );
			}
			while ((cfr==NULL) ? 0 : (clip_end < start))
			{
				NEXT( cfr, clip_start, clip_end );
			}
	
			if (cfr != NULL)
			{
				while ((fr==NULL) ? 0 : (start <= clip_end))
				{
					if ((clip_start <= start) && (clip_end >= end))
					{
						NEXT( fr, start, end );
					}
					else if (clip_start > start)
					{
						if (end < clip_start)
						{
							SETFRONT( start, end );
						}
						else
						{
							SETFRONT( start, clip_start-1 );
						}
						if (clip_end < end)
						{
							start = clip_end + 1;
							NEXT( cfr, clip_start, clip_end );
						}
						else
						{
							NEXT( fr, start, end );
						}
					}
					else /* if (clip_end < end) */
					{
						start = clip_end + 1;
						NEXT( cfr, clip_start, clip_end );
					}
				}
			}
		}
		while (fr != NULL)
		{
			SETFRONT( start, end );
			NEXT( fr, start, end );
		}
/*
		Now draw all the BACK segments of new plane
		-------------------------------------------
			(same algorithm)
*/
		fr = frontier[j];
		cfr = last_frontier[j];
		if (fr != NULL)
		{
			start = fr->ndx;
			end = fr->next->ndx;
		}
		if (cfr != NULL)
		{
			clip_start = cfr->ndx;
			clip_end = cfr->next->ndx;
		}

		while ((fr != NULL) && (cfr != NULL))
		{
			while ((fr==NULL) ? 0 : (end < clip_start))
			{
				SETBACK( start, end );
				NEXT( fr, start, end );
			}
			while ((cfr==NULL) ? 0 : (clip_end < start))
			{
				NEXT( cfr, clip_start, clip_end );
			}
	
			if (cfr != NULL)
			{
				while ((fr==NULL) ? 0 : (start <= clip_end))
				{
					if ((clip_start <= start) && (clip_end >= end))
					{
						NEXT( fr, start, end );
					}
					else if (clip_start > start)
					{
						if (end < clip_start)
						{
							SETBACK( start, end );
						}
						else
						{
							SETBACK( start, clip_start-1 );
						}
						if (clip_end < end)
						{
							start = clip_end + 1;
							NEXT( cfr, clip_start, clip_end );
						}
						else
						{
							NEXT( fr, start, end );
						}
					}
					else /* if (clip_end < end) */
					{
						start = clip_end + 1;
						NEXT( cfr, clip_start, clip_end );
					}
				}
			}
		}
		while (fr != NULL)
		{
			SETBACK( start, end );
			NEXT( fr, start, end );
		}
	}
	swap_frontiers();

} /* BuildPatches */
#else
BuildPatches( int dummy )
{
	int	i, j, k;

	for ( i=ifirst; i<imax; i++ )
	for ( j=jfirst; j<jmax; j++ )
	{
		for ( k=kfirst; k<kmax; )
		{
			if ( (*testFunc)(i,j,k) && !(*testFunc)(i-1,j,k) )
			{
				(*bgnFunc)();
				k = GetSeg_DerInvalide(i,j,k);
				(*endFunc)();
			}
			k++;
	 	}
		for ( k=kfirst; k<kmax; )
		{
			if ( (*testFunc)(i,j,k) && !(*testFunc)(i+1,j,k) )
			{
				(*bgnFunc)();
				k = GetSeg_DevInvalide(i,j,k);
				(*endFunc)();
			}
			k++;
		}
 }
}
#endif

/*--------------------------------------------------------------------------*/
/*			Complete_List
/*--------------------------------------------------------------------------*/
int
Complete_List( int i )
{
 int	jref, kref,
	j, k,
	encore,
	last_axis;
 ListStruct *liste;

 for ( jref=jfirst; jref<jmax; jref++ )
 while ( Seg_GchInvalide[jref] != NULL ) {

	(*bgnFunc)();
	liste = Seg_GchInvalide[jref];
	if ( (Seg_GchInvalide[jref]=liste->next) != NULL ) {
		liste->next->last = NULL;
	}
 	(*setFunc)( LEFT, i, liste->firstj, liste->firstk, liste->dernier_ndx );
	j = jref;
	k = liste->dernier_ndx;
	kref = liste->firstk;
	encore = 1;
	last_axis = K_AXIS;
	while ( encore ) {

		switch ( last_axis ) {
/*
		on debouche d'un segment invalide a gauche
		------------------------------------------
*/
		case K_AXIS:
			if ( !(*testFunc)(i,j,k+1) ) {
				j = GetSeg_SusInvalide(i,j,k);
				last_axis = J_AXIS;
			} else {
				j--; k++;
				j = GetSeg_SousInvalide(i,j,k);
				last_axis = -J_AXIS;
			}
			break;

/*
		On debouche d'un segment invalide a droite
		------------------------------------------
*/
		case -K_AXIS:
			if ( !(*testFunc)(i,j,k-1) ) {
				j = GetSeg_SousInvalide(i,j,k);
				last_axis = -J_AXIS;
			} else {
				j++; k--;
				j = GetSeg_SusInvalide(i,j,k);
				last_axis = J_AXIS;
			}
			break;

/*
		On debouche d'un segment invalide au-dessus
		-------------------------------------------
*/
		case J_AXIS:
			if ( !(*testFunc)(i,j+1,k) ) {
				k = GetSeg_DteInvalide(i,j,k);
				last_axis = -K_AXIS;
			} else {
				j++; k++;
				if ( j==jref && k==kref )
					encore = 0;
				else
				{
					k = GetSeg_GchInvalide(i,j,k);
					last_axis = K_AXIS;
				}
			}
			break;

/*
		On debouche d'un segment invalide au-dessous
		--------------------------------------------
*/
		case -J_AXIS:
			if ( !(*testFunc)(i,j-1,k) ) {
				if ( j==jref && k==kref )
					encore = 0;
				else
				{
					k = GetSeg_GchInvalide(i,j,k);
					last_axis = K_AXIS;
				}
			} else {
				j--; k--;
				k = GetSeg_DteInvalide(i,j,k);
				last_axis = -K_AXIS;
			}
			break;
		}

	}
/*
	La liste est fermee => numpoint+1
*/
#if 0
	(*setFunc)( LEFT, i, liste->firstj, liste->firstk, liste->firstk );
#endif
	(*endFunc)();
 }
}

/*---------------------------------------------------------------------------*/
/*			GetSeg_GchInvalide
/*---------------------------------------------------------------------------*/
static int
GetSeg_GchInvalide( int i, int j, int kref )
{
 ListStruct *liste = Seg_GchInvalide[j];
 int	ndx;

 while (liste->firstk != kref) liste = liste->next;
 if ( liste->last == NULL ) {
	if ( (Seg_GchInvalide[j]=liste->next) != NULL ) {
		liste->next->last = NULL;
	}
 } else {
	liste->last->next = liste->next;
	if ( liste->next != NULL ) liste->next->last = liste->last;
 }
 liste->last = liste->next = NULL;

 ndx = liste->dernier_ndx;
 (*setFunc)( LEFT, i, liste->firstj, liste->firstk, ndx );

 freeList( liste );
 return ndx;
}

/*---------------------------------------------------------------------------*/
/*			GetSeg_DteInvalide
/*---------------------------------------------------------------------------*/
static int
GetSeg_DteInvalide( int i, int j, int kref )
{
 int	k = kref;
 do {
	k--;
 } while ( (*testFunc)(i,j,k) && !(*testFunc)(i,j+1,k) );
 k++;
 (*setFunc)( RIGHT, i, j, kref, k );
 return k;
}

/*---------------------------------------------------------------------------*/
/*			GetSeg_SusInvalide
/*---------------------------------------------------------------------------*/
static int
GetSeg_SusInvalide( int i, int jref, int k )
{
 int	ndx, j=jref;
 do {
	j++;
 } while ( (*testFunc)(i,j,k) && !(*testFunc)(i,j,k+1) );
 j--;
#ifdef OPT
 for (ndx=jref; ndx<=j; ndx++) insert_frontier( ndx, k );
#endif
 (*setFunc)( TOP, i, jref, k, j );
 return j;
}

/*---------------------------------------------------------------------------*/
/*			GetSeg_SousInvalide
/*---------------------------------------------------------------------------*/
static int
GetSeg_SousInvalide( int i, int jref, int k )
{
 int	ndx, j=jref;
 do {
	j--;
 } while ( (*testFunc)(i,j,k) && !(*testFunc)(i,j,k-1) );
 j++;
#ifdef OPT
 for (ndx=j; ndx<=jref; ndx++) insert_frontier( ndx, k );
#endif
 (*setFunc)( BOTTOM, i, jref, k, j );
 return j;
}

/*---------------------------------------------------------------------------*/
/*			GetSeg_DerInvalide
/*---------------------------------------------------------------------------*/
static int
GetSeg_DerInvalide( int i, int j, int kref )
{
 int	k = kref;
 do {
	k++;
 } while ( (*testFunc)(i,j,k) && !(*testFunc)(i-1,j,k) );
 k--;
 (*setFunc)( BACK, i, j, kref, k );
 return k;
}

/*---------------------------------------------------------------------------*/
/*			GetSeg_DevInvalide
/*---------------------------------------------------------------------------*/
static int
GetSeg_DevInvalide( int i, int j, int kref )
{
 int	k = kref;
 do {
	k++;
 } while ( (*testFunc)(i,j,k) && !(*testFunc)(i+1,j,k) );
 k--;
 (*setFunc)( FRONT, i, j, kref, k );
 return k;
}
