 /**************************************************************************
  *                                                                        *
  *               Copyright (C) 1996, Silicon Graphics, Inc.               *
  *                          All Rights Reserved.                          *
  *                                                                        *
  *  The files in this subtree  contain  UNPUBLISHED  PROPRIETARY  SOURCE  *
  *  CODE of Silicon Graphics, Inc.;  the  contents  of  these  files may  *
  *  not be disclosed to third  parties,  copied  or  duplicated  in  any  *
  *  form, in whole or in part,  without  the  prior  written  permission  * 
  *  of  Silicon Graphics, Inc.                                            *
  *                                                                        *
  *  RESTRICTED RIGHTS LEGEND:                                             *
  *  Use,  duplication  or  disclosure  by  the  Government is subject to  *
  *  restrictions as set forth in subdivision (c)(1)(ii) of the Rights in  *
  *  Technical Data and Computer Software clause at  DFARS  252.227-7013,  *
  *  and/or in similar or successor  clauses in the FAR,  DOD or NASA FAR  *
  *  Supplement.  Unpublished - rights reserved  under the Copyright Laws  *
  *  of the United States.                                                 *
  *                                                                        *
  *  THIS SOFTWARE IS  PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,  *
  *  EXPRESS,  IMPLIED OR  OTHERWISE,  INCLUDING  WITHOUT LIMITATION, ANY  *
  *  WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.      *
  *                                                                        *
  *  IN  NO  EVENT  SHALL  SILICON  GRAPHICS  BE  LIABLE FOR ANY SPECIAL,  *
  *  INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF  ANY  KIND,  OR ANY  *
  *  DAMAGES WHATSOEVER RESULTING FROM LOSS  OF  USE,  DATA  OR  PROFITS,  *
  *  WHETHER OR NOT ADVISED OF  THE  POSSIBILITY  OF  DAMAGE,  AND ON ANY  *
  *  THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR  *
  *  PERFORMANCE OF THIS SOFTWARE.                                         *
  **************************************************************************/

/*
    Author : Patrick Bouchaud
             galaad@neu.sgi.com
*/


/*---------------------------------------------------------------------------*/
/*	MACROS, TYPES
/*---------------------------------------------------------------------------*/

#define ADDVEC3(v, v1, v2) v[0]=v1[0]+v2[0]; v[1]=v1[1]+v2[1]; v[2]=v1[2]+v2[2];
#define SUBVEC3(v, v1, v2) v[0]=v1[0]-v2[0]; v[1]=v1[1]-v2[1]; v[2]=v1[2]-v2[2];
#define SETVEC3(v1, v2)    v1[0]=v2[0]; v1[1]=v2[1]; v1[2]=v2[2];
#define SCLVEC3(v1, x)     v1[0]*=x; v1[1]*=x; v1[2]*=x;

#define NRMVEC3(v) {\
    float norm = fsqrt(v[0]*v[0]+v[1]*v[1]+v[2]*v[2]);\
    v[0]/=norm; v[1]/=norm; v[2]/=norm;\
}

#define LB(plane,i)   (plane.farCentre[(i)] - plane.up[(i)])
#define LT(plane,i)   (plane.farCentre[(i)] + plane.dn[(i)])
#define RB(plane,i)   (plane.farCentre[(i)] - plane.dn[(i)])
#define RT(plane,i)   (plane.farCentre[(i)] + plane.up[(i)])

typedef struct _vertex {
	float x, y, z;
	struct _vertex *next, *last;
} Vertex;

typedef struct {
	GLdouble Equation[4];
	GLfloat	up[3], dn[3],
			S[3], T[3],
			farCentre[3];
} PlaneStructure;

int updatePlaneStructure( PlaneStructure * );

/*---------------------------------------------------------------------------*/
/*	GLOBAL data
/*---------------------------------------------------------------------------*/
#ifdef MAIN_PROGRAM

PlaneStructure Plane;

/*---------------------------------------------------------------------------*/
/*	EXTERN references
/*---------------------------------------------------------------------------*/
#else

extern PlaneStructure Plane;

#endif
