 /**************************************************************************
  *                                                                        *
  *               Copyright (C) 1996, Silicon Graphics, Inc.               *
  *                          All Rights Reserved.                          *
  *                                                                        *
  *  The files in this subtree  contain  UNPUBLISHED  PROPRIETARY  SOURCE  *
  *  CODE of Silicon Graphics, Inc.;  the  contents  of  these  files may  *
  *  not be disclosed to third  parties,  copied  or  duplicated  in  any  *
  *  form, in whole or in part,  without  the  prior  written  permission  * 
  *  of  Silicon Graphics, Inc.                                            *
  *                                                                        *
  *  RESTRICTED RIGHTS LEGEND:                                             *
  *  Use,  duplication  or  disclosure  by  the  Government is subject to  *
  *  restrictions as set forth in subdivision (c)(1)(ii) of the Rights in  *
  *  Technical Data and Computer Software clause at  DFARS  252.227-7013,  *
  *  and/or in similar or successor  clauses in the FAR,  DOD or NASA FAR  *
  *  Supplement.  Unpublished - rights reserved  under the Copyright Laws  *
  *  of the United States.                                                 *
  *                                                                        *
  *  THIS SOFTWARE IS  PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,  *
  *  EXPRESS,  IMPLIED OR  OTHERWISE,  INCLUDING  WITHOUT LIMITATION, ANY  *
  *  WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.      *
  *                                                                        *
  *  IN  NO  EVENT  SHALL  SILICON  GRAPHICS  BE  LIABLE FOR ANY SPECIAL,  *
  *  INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF  ANY  KIND,  OR ANY  *
  *  DAMAGES WHATSOEVER RESULTING FROM LOSS  OF  USE,  DATA  OR  PROFITS,  *
  *  WHETHER OR NOT ADVISED OF  THE  POSSIBILITY  OF  DAMAGE,  AND ON ANY  *
  *  THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR  *
  *  PERFORMANCE OF THIS SOFTWARE.                                         *
  **************************************************************************/

/*
    Author : Patrick Bouchaud
             galaad@neu.sgi.com
*/

#define		MAIN_PROGRAM

#include	<bstring.h>
#include	<stdio.h>
#include	<sys/types.h>
#include	<sys/time.h>
#include	<sys/signal.h>
#include	<sys/stat.h>
#include	<unistd.h>
#include	<fcntl.h>
#include	<malloc.h>

#include    <GL/gl.h>
#include    <GL/glut.h>

#include	"envmap.h"

static int size = 256;

static float lightposition[] = { 1., 1., 0., 0. };
static float rotmat[] =
{
	1.,0.,0.,0.,
	0.,1.,0.,0.,
	0.,0.,1.,0.,
	0.,0.,0.,1.
};

static struct {
	int		lastx, lasty,
			middledown, rightdown, leftdown;
	float	dx, dy;
} mouse;

static int	envmapWin, W, H;
static float Rx=0., Ry=0.;

static void Reshape( int w, int h );
static void Redraw( void );
static void Mouse( int, int, int, int );
static void Motion( int, int );
static void Key( unsigned char, int, int );

static void InitGraphique( void );

#ifdef SAVEIMG
#define SAVEFACE(name,face) fromcpack((char *)face, FACESIZE, FACESIZE, name);
#else
#define SAVEFACE( name, face )
#endif

#define DRAWSCENE()\
	glEnable( GL_LIGHT0 );\
	glLightfv( GL_LIGHT0, GL_POSITION, lightposition );\
	draw_logo();

/*===========================================================================*/
/*				main
/*===========================================================================*/
int
main( int argc, char *argv[] )
{
	glutInit( &argc, argv );

	glutInitDisplayMode( GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH );
	glutInitWindowPosition( 100, 100 );
	glutInitWindowSize( 900, 900 );
	envmapWin = glutCreateWindow( "Self-Generated Environment Map" );
	glutDisplayFunc( Redraw );
	glutReshapeFunc( Reshape );
	glutMouseFunc( Mouse );
	glutMotionFunc( Motion );
	glutKeyboardFunc( Key );
	glutIdleFunc( Redraw );

	glutMainLoop();
}

/*---------------------------------------------------------------------------*/
/*      Reshape
/*---------------------------------------------------------------------------*/
static void
Reshape( int w, int h )
{
	static int first = 1;
	glViewport( 0, 0, w, h );
	W = w;
	H = h;

	if (first)
	{
		InitGraphique();
		init_cube();
		first = 0;
	}
}

/*---------------------------------------------------------------------------*/
/*	Mouse
/*---------------------------------------------------------------------------*/
static void Mouse( int button, int state, int x, int y )
{
	mouse.lastx = x;
	mouse.lasty = y;
	mouse.dx = 0.;
	mouse.dy = 0.;
	switch (button)
	{
		case GLUT_LEFT_BUTTON:
			mouse.leftdown = (state==GLUT_DOWN);
			break;

		case GLUT_RIGHT_BUTTON:
			mouse.rightdown = (state==GLUT_DOWN);
			break;

		case GLUT_MIDDLE_BUTTON:
			mouse.middledown = (state==GLUT_DOWN);
			break;
	}
	glutPostRedisplay();
}

/*---------------------------------------------------------------------------*/
/*	Motion
/*---------------------------------------------------------------------------*/
static void
Motion( int x, int y )
{
	mouse.dx = x-mouse.lastx;
	mouse.dy = y-mouse.lasty;
	mouse.lastx = x;
	mouse.lasty = y;

	glutPostRedisplay();
}

/*---------------------------------------------------------------------------*/
/*	Key
/*---------------------------------------------------------------------------*/
static void
Key( unsigned char key, int x, int y )
{
	switch (key)
	{
		case 27:	/* Escape */
			exit(0);

		case 'z':
		case 'Z':
			switch (size)
			{
				case 256:
					size = 512;
					break;

				case 512:
					size = 256;
					break;
			}
			bzero( envmap, size*size*sizeof(long) );
			break;
	}
	glutPostRedisplay();
}


/*---------------------------------------------------------------------------*/
/*	Redraw
/*---------------------------------------------------------------------------*/
static void
Redraw()
{
	float envmat[16];

	if (mouse.middledown)
	{
		Rx = -mouse.dy * .1;
		Ry = -mouse.dx * .1;
	}

	glViewport( 0, 0, W, H );
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	gluPerspective( 60., 1., .1, 100. );
	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();

/*
	Matrix computation
	------------------
*/
	glLoadIdentity();
	glTranslatef( 0., 0., -5. );

	glPushMatrix();
	glLoadIdentity();
	glRotatef( Ry, 0., 1., 0. );
	glRotatef( Rx, 1., 0., 0. );
	glMultMatrixf( rotmat );
	glGetFloatv( GL_MODELVIEW_MATRIX, rotmat );
	glPopMatrix();

	glMultMatrixf( rotmat );

/*
	We only update the rotation matrix
	----------------------------------
*/
	glGetFloatv( GL_MODELVIEW_MATRIX, envmat );
	envmat[12] = envmat[13] = envmat[14] = 0.;

/*
		Drawing
		-------
*/
	glClearColor( .5, .5, 1., 1. );
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

	DRAWSCENE();

/*
	======================================================================

		We could re-read all the faces (see init_cube()), and re-generate
	the environment map using genenvmap() with an Identity matrix
	instead of envmat. BUT we know the faces have only been rotated.
	In real applications it is often the case that the environment
	of an object is just rotated relatively to the object, rather than
	deformed (BTW, light sources have to stay fixed also, in this case).

		If you're using video acquisition to update one or part of one of
	your faces, and your environment is steady, then you don't even need
	to rotate the faces - in this case, better directly replace the
	relevant pixels in your envmap.

		At this stage, we just draw the environment map.
	In a real application, we would load it into the Texture memory
	with glTexImage2D(), and use it as a GL_SPHERE_MAP with glTexGen()
	on any object located inside of this environment

	======================================================================
*/
	
	genenvmap( envmat, size );

	glViewport( 0, 0, W, H );
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	glOrtho( 0., (GLdouble)W, 0., (GLdouble)H, -1., 1. );
	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();

	glAlphaFunc( GL_NOTEQUAL, 0 );
	glEnable( GL_ALPHA_TEST );
	glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
	glRasterPos2i( (W-size)/2, (H-size)/2 );
	glDrawPixels( size, size, GL_ABGR_EXT, GL_UNSIGNED_BYTE, envmap );
	glDisable( GL_ALPHA_TEST );

	glutSwapBuffers();
}

/*---------------------------------------------------------------------------*/
/*                      InitGraphique
/*---------------------------------------------------------------------------*/
static void
InitGraphique( void )
{
	float lightAmbient[] = {0., 0., 0., 0.};
	float diffuseColor[] = { 1., 1., 1., 1. };
	float ambientColor[] = { 0.0, 0.0, 0.2, 1. };
	float specularColor[] = { 0.1, 0.1, 0.5, 1. };
	float emissiveColor[] = { 0.0, 0.0, 0.0, 1. };
	float shininess = .1;

/*
	Fenetre de Visualisation
	------------------------
*/
	glEnable( GL_DEPTH_TEST );
	glDepthFunc( GL_LEQUAL );

	glShadeModel( GL_SMOOTH );
	glClearColor( 0., 0., 0., 1. );
	glClearDepth( 1. );
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

	glLightfv( GL_LIGHT0, GL_DIFFUSE, diffuseColor );
	glLightfv( GL_LIGHT0, GL_AMBIENT, diffuseColor );

	glLightModeli( GL_LIGHT_MODEL_LOCAL_VIEWER, 0 );
	glLightModelfv( GL_LIGHT_MODEL_AMBIENT, lightAmbient );

    glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT, ambientColor );
    glMaterialfv( GL_FRONT_AND_BACK, GL_DIFFUSE, diffuseColor );
    glMaterialfv( GL_FRONT_AND_BACK, GL_SPECULAR, specularColor );
    glMaterialfv( GL_FRONT_AND_BACK, GL_EMISSION, emissiveColor );
    glMaterialf ( GL_FRONT_AND_BACK, GL_SHININESS, shininess );

	glEnable( GL_LIGHTING );

} /* InitGraphique */

/*---------------------------------------------------------------------------*/
/*	init_cube
/*---------------------------------------------------------------------------*/
init_cube()
{
	glViewport( 0, 0, FACESIZE, FACESIZE );
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	gluPerspective( 90., 1., ENVNEAR, ENVFAR );
	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();

	glReadBuffer( GL_BACK );
	glDrawBuffer( GL_BACK );
	glClearColor( 0., 0., .5, 1. );

#define	GETFACE( angle, xaxis, yaxis, zaxis, face, file )\
	\
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );\
	glPushMatrix();\
	glRotatef( angle, xaxis, yaxis, zaxis );\
	DRAWSCENE();\
	glReadPixels(\
		0, 0, FACESIZE, FACESIZE,\
		GL_ABGR_EXT,GL_UNSIGNED_BYTE,\
		face\
	);\
	glPopMatrix();\
	SAVEFACE( file, face );

	GETFACE(    0., 0., 1., 0., front_face, "front.rgba" );
	GETFACE(  -90., 0., 1., 0., right_face, "right.rgba" );
	GETFACE( -180., 0., 1., 0., back_face,  "back.rgba" );
	GETFACE( -270., 0., 1., 0., left_face,  "left.rgba" );
	GETFACE(   90., 1., 0., 0., top_face,   "top.rgba" );
	GETFACE(  -90., 1., 0., 0., bottom_face, "bottom.rgba" );
}
