 /**************************************************************************
  *                                                                        *
  *               Copyright (C) 1996, Silicon Graphics, Inc.               *
  *                          All Rights Reserved.                          *
  *                                                                        *
  *  The files in this subtree  contain  UNPUBLISHED  PROPRIETARY  SOURCE  *
  *  CODE of Silicon Graphics, Inc.;  the  contents  of  these  files may  *
  *  not be disclosed to third  parties,  copied  or  duplicated  in  any  *
  *  form, in whole or in part,  without  the  prior  written  permission  * 
  *  of  Silicon Graphics, Inc.                                            *
  *                                                                        *
  *  RESTRICTED RIGHTS LEGEND:                                             *
  *  Use,  duplication  or  disclosure  by  the  Government is subject to  *
  *  restrictions as set forth in subdivision (c)(1)(ii) of the Rights in  *
  *  Technical Data and Computer Software clause at  DFARS  252.227-7013,  *
  *  and/or in similar or successor  clauses in the FAR,  DOD or NASA FAR  *
  *  Supplement.  Unpublished - rights reserved  under the Copyright Laws  *
  *  of the United States.                                                 *
  *                                                                        *
  *  THIS SOFTWARE IS  PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,  *
  *  EXPRESS,  IMPLIED OR  OTHERWISE,  INCLUDING  WITHOUT LIMITATION, ANY  *
  *  WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.      *
  *                                                                        *
  *  IN  NO  EVENT  SHALL  SILICON  GRAPHICS  BE  LIABLE FOR ANY SPECIAL,  *
  *  INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF  ANY  KIND,  OR ANY  *
  *  DAMAGES WHATSOEVER RESULTING FROM LOSS  OF  USE,  DATA  OR  PROFITS,  *
  *  WHETHER OR NOT ADVISED OF  THE  POSSIBILITY  OF  DAMAGE,  AND ON ANY  *
  *  THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR  *
  *  PERFORMANCE OF THIS SOFTWARE.                                         *
  **************************************************************************/

/*
    Author : Patrick Bouchaud
             galaad@neu.sgi.com
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "envmap.h"

static int init_vectors( int size, int *tmin, int *tmax, float *vectors );
static int update_pixels( float mat[4][4], int size, int *tmin, int *tmax, float *vectors );
static unsigned long setpixel( register float v[3] );
static unsigned long interpol( unsigned long *, float, float );

static float vectors_256[3*256*256];
static int tmin_256[256], tmax_256[256];

static float vectors_512[3*512*512];
static int tmin_512[512], tmax_512[512];

/*---------------------------------------------------------------------------*/
/*	genenvmap
/*---------------------------------------------------------------------------*/
genenvmap( mat, size )
float mat[4][4];
int size;
{
 static int first = 1;

 if (first)
 {
	first = 0;
	init_vectors( 256, tmin_256, tmax_256, vectors_256 );
	init_vectors( 512, tmin_512, tmax_512, vectors_512 );
 }
 switch (size)
 {
	case 256:
		update_pixels( mat, 256, tmin_256, tmax_256, vectors_256 );
		break;
	case 512:
		update_pixels( mat, 512, tmin_512, tmax_512, vectors_512 );
		break;
 }

}

/*---------------------------------------------------------------------------*/
/*	init_vectors
/*---------------------------------------------------------------------------*/
static int
init_vectors( int size, int *tmin, int *tmax, float *vectors )
{
	float delta = 1./(float)(size-1);

	int		s;
	float	fs, *v0;

	for (s=0, fs=-.5; s<size; s++, fs+=delta)
	{
		int		t = 0;
		float	aux,
				ft = -.5,
				ftmin = .25 - fs*fs;

		while ((ft*ft > ftmin) && (t<size))
		{
			t++; ft+=delta;
		}

		tmin[s] = t;
		tmax[s] = size - t;
		v0 = &vectors[ 3*(tmin[s]*size + s) ];
		for (; t<tmax[s]; t++, ft+=delta, v0+=3*size)
		{
			v0[2] = 1. - 8.*(fs*fs + ft*ft);
			aux = fsqrt(8.*(v0[2]+1.));
			v0[0] = aux*fs;
			v0[1] = aux*ft;
		}
	}
}

/*---------------------------------------------------------------------------*/
/*	update_pixels
/*---------------------------------------------------------------------------*/
static int
update_pixels( float mat[4][4], int size, int *tmin, int *tmax, float *vectors )
{
	float invmat[16];
	int	s, t, col;

	mat_invert( mat, invmat );

	for (s=0; s<size; s++)
	{
		register float	v1[3];
		int		ndx = s*size + tmin[s];
		float	*v0 = &vectors[ 3*ndx ];
		unsigned long *map = &envmap[ndx];

		for (t=tmin[s]; t<tmax[s]; t++, v0+=3)
 		{
			v1[0] = invmat[0]*v0[0] + invmat[4]*v0[1] + invmat[8]*v0[2];
			v1[1] = invmat[1]*v0[0] + invmat[5]*v0[1] + invmat[9]*v0[2];
			v1[2] = invmat[2]*v0[0] + invmat[6]*v0[1] + invmat[10]*v0[2];

			*map++ = setpixel( v1 );
		}
	}
}

/*---------------------------------------------------------------------------*/
/*	setpixel
/*---------------------------------------------------------------------------*/
static unsigned long
setpixel(v)
register float v[3];
{
 int	i, facenum, xint, yint;
 float	x, y, max, scale = HALFSIZE;

 if (v[0] > 0)
 {
	max = v[0];
	facenum = LEFT_FACE;
 }
 else
 {
	max = -v[0];
	facenum = RIGHT_FACE;
 }
 if (v[1] > max)
 {
	max = v[1];
	facenum = BOTTOM_FACE;
 }
 else if (v[1] < -max)
 {
	max = -v[1];
	facenum = TOP_FACE;
 }
 if (v[2] > max)
 {
	max = v[2];
	facenum = BACK_FACE;
 }
 else if (v[2] < -max)
 {
	max = -v[2];
	facenum = FRONT_FACE;
 }

#ifndef INTERPOL
#define	SETPIXEL( f, x, y )\
 if ((xint=(int)x) > XYMAX) xint = XYMAX;\
 if ((yint=(int)y) > XYMAX) yint = XYMAX;\
 return f[(yint<<FACEBIT) + xint];
#else
#define	SETPIXEL( f, x, y )\
	return interpol( f, x, y );
#endif

 switch (facenum)
 {
	case BACK_FACE:
		scale = HALFSIZE/v[2];
		x = (v[2] - v[0]) * scale;
		y = (v[2] + v[1]) * scale;
		SETPIXEL( back_face, x, y );
		break;

	case FRONT_FACE:
		scale = HALFSIZE/v[2];
		x = (v[2] - v[0]) * scale;
		y = (v[2] - v[1]) * scale;
		SETPIXEL( front_face, x, y );
		break;

	case RIGHT_FACE:
		scale = HALFSIZE/v[0];
		x = (v[0] + v[2]) * scale;
		y = (v[0] - v[1]) * scale;
		SETPIXEL( right_face, x, y );
		break;

	case LEFT_FACE:
		scale = HALFSIZE/v[0];
		x = (v[0] + v[2]) * scale;
		y = (v[0] + v[1]) * scale;
		SETPIXEL( left_face, x, y );
		break;

	case TOP_FACE:
		scale = HALFSIZE/v[1];
		x = (v[1] - v[0]) * scale;
		y = (v[1] + v[2]) * scale;
		SETPIXEL( top_face, x, y );
		break;

	case BOTTOM_FACE:
		scale = HALFSIZE/v[1];
		x = (v[1] + v[0]) * scale;
		y = (v[1] + v[2]) * scale;
		SETPIXEL( bottom_face, x, y );
		break;
 }
}

/*---------------------------------------------------------------------------*/
/*	interpol
/*---------------------------------------------------------------------------*/
static unsigned long
interpol( unsigned long *f, float x, float y )
{
 int	i,
		x0 = INF((int)x, XYMAX),
		y0 = INF((int)y, XYMAX),
		x1 = INF(x0+1, XYMAX),
		y1 = INF(y0+1, XYMAX);
 float	tx = x - (float) x0,
		ty = y - (float) y0,
		cp[4];

 union {
	unsigned long packed;
	char cp[4];
 } pix00, pix01, pix10, pix11, pixel;

#define pixel(x,y) f[(y)*FACESIZE + (x)]

 pix00.packed = pixel(x0,y0);
 pix01.packed = pixel(x0,y1);
 pix10.packed = pixel(x1,y0);
 pix11.packed = pixel(x1,y1);

 for (i=0; i<4; i++)
 {
	cp[i] =
		(float) pix00.cp[i] * (1.-tx)*(1.-ty) +
		(float) pix10.cp[i] * tx * (1.-ty) +
		(float) pix01.cp[i] * (1.-tx) * ty +
		(float) pix11.cp[i] * tx * ty;
	pixel.cp[i] = cp[i];
 }

 return pixel.packed;
}
